"""
Instrument driver for Thorlabs TC200 temperature controller.
"""

from enum import Enum
import logging
from typing import Any, NamedTuple, Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport


# Logger for this module
_logger = logging.getLogger(__name__)


# Default timeout for command responses
TIMEOUT = 1


# Various limits
TMIN = 20.0
TMAX = 200.0
TLIMIT_MIN = 20.0
TLIMIT_MAX = 205.0
PLIMIT_MIN = 0.1
PLIMIT_MAX = 18.0
PGAIN_MIN = 1
PGAIN_MAX = 250
IGAIN_MIN = 0
IGAIN_MAX = 250
DGAIN_MIN = 0
DGAIN_MAX = 250


# Error responses
CMD_ERROR = "CMD_NOT_DEFINED"
ARG_RANGE_ERROR = "CMD_ARG_RANGE_ERR"


# Bitmasks for sensor bits
MASK_OUTPUT_STATE   = 0x01
MASK_MODE           = 0x02
MASK_SENSOR_TYPE    = 0x0C
MASK_DISPLAY_UNIT   = 0x30
MASK_ALARM          = 0x40
MASK_CYCLE_STATE    = 0x80


class _Command(Enum):
    """Commands."""
    TOGGLE_ENABLE = "ens"
    SET_TEMPERATURE = "tset"
    PGAIN = "pgain"
    IGAIN = "igain"
    DGAIN = "dgain"
    MAX_POWER = "pmax"  # user manual incorrectly uses PMAX
    MAX_TEMPERATURE = "tmax"  # user manual incorrectly uses TMAX
    TUNE_OFFSET = "tune"


class _Query(Enum):
    """Queries."""
    SETPOINT_TEMPERATURE = "tset?"
    ACTUAL_TEMPERATURE = "tact?"
    STATUS = "stat?"
    PID_GAINS = "pid?"
    MAX_POWER = "pmax?"  # user manual incorrectly uses PMAX?
    MAX_TEMPERATURE = "tmax?"  # user manual incorrectly uses TMAX?
    TUNED = "tune?"
    IDENT = "id?"


class OutputState(Enum):
    """Output state (enabled/disabled)."""
    DISABLED = 0x00
    ENABLED = 0x01


class Mode(Enum):
    """Control mode (constant temperature/cycling)."""
    NORMAL = 0x00
    CYCLE = 0x02


class SensorType(Enum):
    """Sensor type enumeration."""
    TH10K = 0x00
    PTC100 = 0x04
    PTC1000 = 0x08


class DisplayUnit(Enum):
    """Front panel display units."""
    KELVIN = 0x00
    CELSIUS = 0x10
    FAHRENHEIT = 0x20


class SensorAlarmState(Enum):
    """Sensor alarm state."""
    NO_ALARM = 0x00
    ALARM = 0x40


class CycleState(Enum):
    """Cycle state."""
    NOT_PAUSED = 0x00
    PAUSED = 0x80


class Tc200Status(NamedTuple):
    """TC 200 status.

    Attributes:
        output_state:   Output state (enabled/disabled).
        mode:           Output mode (normal/cycle).
        sensor_type:    Sensor type.
        unit:           Display unit (C/F/K).
        alarm:          Sensor alarm state (alarm/no alarm).
        cycle_state:    Cycle state (paused/not paused).
    """
    output_state: OutputState
    mode: Mode
    sensor_type: SensorType
    unit: DisplayUnit
    alarm: SensorAlarmState
    cycle_state: CycleState


class Thorlabs_TC200(QMI_Instrument):
    """Instrument driver for the Thorlabs TC200 temperature controller."""

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        A subset of the commands and queries of the TC200 is implmented. Specifically, configuring the temperature
        (typically done only once) and cycling is not implemented.

        The device uses an FTDI USB-to-serial interface at 115200 baud (8 data bits, no parity, 1 stop bit, no flow
        control). When connected directly to the device, use `serial:<device>:baudrate=115200` as transport
        specification.

        Parameters:
            context:    QMI Context.
            name:       Identifier for the instrument in the QMI context.
            transport:  Transport specification string.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200})

    def _send_command(self, command: _Command, argument: Any = None) -> str:
        self._check_is_open()

        if argument is None:
            command_string = f"{command.value}\r"
        else:
            command_string = f"{command.value}={argument}\r"

        _logger.debug("TC200 command: %s", command_string.strip())
        self._transport.write(command_string.encode("utf-8"))
        response = self._receive_response()
        if CMD_ERROR in response:
            raise QMI_InstrumentException("Command not accepted {}".format(command_string))

        return response

    def _send_query(self, query: _Query) -> str:
        self._check_is_open()

        query_string = f"{query.value}\r"

        _logger.debug("TC200 query: %s", query_string.strip())
        self._transport.write(query_string.encode("utf-8"))
        response = self._receive_response()
        if CMD_ERROR in response:
            raise QMI_InstrumentException("Query not accepted {}".format(query_string))

        return response

    def _receive_response(self) -> str:
        # The device echoes the command, followed by carriage return ('\r'), the actual response, another carriage
        # return and then a "prompt" ("> "). All responses are a single line of text, except the "config?" query, which
        # spans multiple lines separated by carriage returns. (Due to the use of carriage returns, miniterms on a Unix
        # environment do not work well).
        try:
            response = self._transport.read_until(b"\r> ", timeout=TIMEOUT)
            parts = response.decode('ascii').split('\r')
        except UnicodeDecodeError as exc:
            raise QMI_InstrumentException("Invalid instrument response") from exc

        # parts[0] is the echoed command, parts[-1] is the new "prompt"; multiline responses are joined by '\n' (single
        # line responses become a normal string).
        response_string = '\n'.join(parts[1:-1])
        _logger.debug('TC200 response: %s', response_string.strip())
        return response_string

    def _get_status(self) -> Tc200Status:
        self._check_is_open()
        self._transport.write(b"stat?\r")

        # The status query returns a slightly different response than other queries: it returns the hexadecimal
        # representation of the status byte, without the '\r' that terminates other response strings. The hexadecimal
        # is printed without leading zero, so it can be either one or two characters.
        try:
            response = self._transport.read_until(b" > ", timeout=TIMEOUT)
            status_string = response.decode('ascii')[len("stat?\r"):-len(" > ")]  # expected: "stat?\r.. > "
            status_byte = int(status_string, base=16)
            status = Tc200Status(
                output_state=OutputState(status_byte & MASK_OUTPUT_STATE),
                mode=Mode(status_byte & MASK_MODE),
                sensor_type=SensorType(status_byte & MASK_SENSOR_TYPE),
                unit=DisplayUnit(status_byte & MASK_DISPLAY_UNIT),
                alarm=SensorAlarmState(status_byte & MASK_ALARM),
                cycle_state=CycleState(status_byte & MASK_CYCLE_STATE)
            )
        except (UnicodeDecodeError, ValueError) as exc:
            raise QMI_InstrumentException("Invalid status response") from exc

        return status

    def _set_gain(self, cmd: _Command, gain: int, min_limit: int, max_limit: int) -> None:
        gain = int(gain)
        if min_limit <= gain <= max_limit:
            gain_string = f"{gain:d}"
        else:
            raise QMI_UsageException(
                "Gain must be between {} and {} inclusive, got {}".format(min_limit, max_limit, gain)
            )

        response = self._send_command(cmd, gain_string)
        if ARG_RANGE_ERROR in response:
            raise QMI_InstrumentException("Gain value not accepted: {}".format(gain_string))

    @rpc_method
    def open(self) -> None:
        """Open the device interface."""
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        """Close the device interface."""
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Retrieve the device identification."""
        ident = self._send_query(_Query.IDENT)
        vendor, model, _, version = ident.split(' ')
        return QMI_InstrumentIdentification(vendor=vendor, model=model, serial="", version=version)

    @rpc_method
    def get_status(self) -> Tc200Status:
        """Retrieve and parse device status."""
        return self._get_status()

    @rpc_method
    def is_enabled(self) -> bool:
        """Determine if the heater output is enabled."""
        return self._get_status().output_state == OutputState.ENABLED

    @rpc_method
    def has_alarm(self) -> bool:
        """Determine if sensor alarm is set."""
        return self._get_status().alarm == SensorAlarmState.ALARM

    @rpc_method
    def is_tuned(self) -> bool:
        """Determine if the offset calibration has been performed."""
        response = self._send_query(_Query.TUNED)

        # When response = "0", it means no offset tuning has been performed. A nonzero response should give the offset
        # value (according to the user manual), but it is unclear what the units are.
        return int(response) > 0

    @rpc_method
    def enable(self) -> None:
        """Enable the heater output."""
        if self.has_alarm():
            raise QMI_InstrumentException("Heater cannot be enabled due to sensor alarm")

        if not self.is_enabled():
            self._send_command(_Command.TOGGLE_ENABLE)
        else:
            _logger.info("Ignored enable command: heater already enabled")

    @rpc_method
    def disable(self) -> None:
        """Disable the heater output."""
        if self.is_enabled():
            self._send_command(_Command.TOGGLE_ENABLE)
        else:
            _logger.info("Ignored disable command: heater already disabled")

    @rpc_method
    def set_temperature(self, setpoint: float) -> None:
        """Set the temperature.

        If `setpoint` exceeds the TMAX limit or 200.0 deg Celsius, whichever is lower, the setpoint will be refused by
        the device. Minimum temperature setting is 20.0 deg Celsius. Resolution is 0.1 deg.

        Parameters:
            setpoint:   temperature setpoint in degrees Celsius.
        """
        if TMIN <= setpoint <= TMAX:
            setpoint_string = f"{setpoint:.1f}"
        else:
            raise QMI_UsageException(
                "Temperature setpoint must be between {} and {} inclusive, got {}".format(TMIN, TMAX, setpoint)
            )

        response = self._send_command(_Command.SET_TEMPERATURE, setpoint_string)
        if ARG_RANGE_ERROR in response:
            raise QMI_InstrumentException("Setpoint not accepted: {}".format(setpoint_string))

    @rpc_method
    def get_temperature_setpoint(self) -> float:
        """Get the current temperature setpoint in degrees Celsius."""
        response = self._send_query(_Query.SETPOINT_TEMPERATURE)
        value = response.split()[0]  # response is "XX.X Celsius"
        return float(value)

    @rpc_method
    def get_temperature_actual(self) -> float:
        """Get the actual temperature in degrees Celsius."""
        response = self._send_query(_Query.ACTUAL_TEMPERATURE)
        value = response.split()[0]  # response is "XX.X C"
        return float(value)

    @rpc_method
    def set_temperature_limit(self, limit: float) -> None:
        """Set the upper temperature setpoint limit.

        Parameters:
            limit:  upper temperature limit in degrees Celsius.
        """
        if TLIMIT_MIN <= limit <= TLIMIT_MAX:
            limit_string = f"{limit:.1f}"
        else:
            raise QMI_UsageException(
                "Temperature limit must be between {} and {} inclusive, got {}".format(TLIMIT_MIN, TLIMIT_MAX, limit)
            )

        response = self._send_command(_Command.MAX_TEMPERATURE, limit_string)
        if ARG_RANGE_ERROR in response:
            raise QMI_InstrumentException("Limit value not accepted: {}".format(limit_string))

    @rpc_method
    def get_temperature_limit(self) -> float:
        """Get the upper temperature setpoint limit in degrees Celsius."""
        response = self._send_query(_Query.MAX_TEMPERATURE)
        return float(response)  # response is "XX.X"

    @rpc_method
    def set_power_limit(self, limit: float) -> None:
        """Set the upper power limit.

        Parameters:
            limit:  upper power limit in Watt.
        """
        if PLIMIT_MIN <= limit <= PLIMIT_MAX:
            limit_string = f"{limit:.1f}"
        else:
            raise QMI_UsageException(
                "Power limit must be between {} and {} inclusive, got {}".format(PLIMIT_MIN, PLIMIT_MAX, limit)
            )

        response = self._send_command(_Command.MAX_POWER, limit_string)
        if ARG_RANGE_ERROR in response:
            raise QMI_InstrumentException("Limit value not accepted: {}".format(limit_string))

    @rpc_method
    def get_power_limit(self) -> float:
        """Get the upper power limit in Watt."""
        response = self._send_query(_Query.MAX_POWER)
        return float(response)  # response is "XX.X"

    @rpc_method
    def set_pid_gains(self,
                      p_gain: Optional[int] = None,
                      i_gain: Optional[int] = None,
                      d_gain: Optional[int] = None
                     ) -> None:
        """Set P, I, D gains for the control loop.

        Parameters:
            p_gain:  proportional gain between 1 and 250 inclusive.
            i_gain:  integral gain between 0 and 250 inclusive.
            d_gain:  differential gain between 0 and 250 inclusive.
        """
        if p_gain is not None:
            self._set_gain(_Command.PGAIN, p_gain, PGAIN_MIN, PGAIN_MAX)

        if i_gain is not None:
            self._set_gain(_Command.IGAIN, i_gain, IGAIN_MIN, IGAIN_MAX)

        if d_gain is not None:
            self._set_gain(_Command.DGAIN, d_gain, DGAIN_MIN, DGAIN_MAX)

    @rpc_method
    def get_pid_gains(self) -> Tuple[int, int, int]:
        """Get P, I, D gains for the control loop."""
        response = self._send_query(_Query.PID_GAINS)
        p_gain, i_gain, d_gain = [int(x) for x in response.split()]  # response is X Y Z
        return p_gain, i_gain, d_gain

    @rpc_method
    def tune(self) -> None:
        """Tune the temperature offset.

        The device can automatically tune the temperature offset. This works best if the P-gain is set to a moderate
        value (around 100-125) and the I-gain and D-gain are both zero. Tuning should *only* be performed when the
        temperature has reached a steady state, otherwise the results are unpredictable (note that as a result you
        cannot just `detune()` immediately followed by `tune()`, you need to wait for stead state after the detune).

        The heater output must be enabled and the device must not be already tuned.

        Tuning takes some time and the temperature may overshoot the setpoint a bit, depending on he P-gain. Typical
        time scale is about a minute or more. The final temperature should be within 0.1 degree Celsius.

        The offset is stored in persistent memory and loaded automatically on device startup.
        """
        if not self.is_enabled():
            raise QMI_UsageException("Heater output must be enabled for tuning")

        if self.is_tuned():
            raise QMI_UsageException("Device is already tuned")

        self._send_command(_Command.TUNE_OFFSET)

    @rpc_method
    def detune(self) -> None:
        """Detune (remove) the temperature offset.

        Allow the temperature to settle before re-tuning!
        """
        if self.is_tuned():
            # Issuing a tune command when the device is already tuned results in a detune.
            self._send_command(_Command.TUNE_OFFSET)
        else:
            _logger.debug("Device offset is not tuned, ignoring request for detune")



"""Module containing the packets for the APT protocol."""

from typing import ClassVar
from qmi.instruments.thorlabs.apt_protocol import (
    AptMessage,
    AptMessageId,
    apt_long,
    apt_dword,
    apt_char,
    apt_word,
    apt_byte,
    apt_short,
)


class HW_GET_INFO(AptMessage):
    """
    Data packet structure for the HW_GET_INFO response. This packet is sent as a response to HW_GET_INFO.

    Fields:
        serial_number:      Serial number of device.
        model_number:       Model number of device.
        type:               Hardware type of device.
        firmware_version:   Firmware version of device.
        hw_version:         Hardware version of device.
        mod_state:          Modification state of device.
        num_channels:       Number of channels in device.
    """

    MESSAGE_ID = AptMessageId.HW_GET_INFO.value
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("serial_number", apt_long),
        ("model_number", apt_char * 8),
        ("type", apt_word),
        ("firmware_version", apt_dword),
        (
            "internal",
            apt_dword * 15,
        ),  # this is for internal use, so we don't know what type it returns
        ("hw_version", apt_word),
        ("mod_state", apt_word),
        ("num_channels", apt_word),
    ]


class MOD_GET_CHANENABLESTATE(AptMessage):
    """
    Header structure for the MOD_GET_CHANENABLESTATE response. This header is sent as a response to
    MOD_REQ_CHANENABLESTATE.

    Fields:
        message_id:     ID of message.
        chan_ident:     Channel number.
        enable_state:   Indicate whether chanel is enabled or disabled.
        dest:           Destination of message.
        source:         Source of message.
    """

    MESSAGE_ID = AptMessageId.MOD_GET_CHANENABLESTATE.value
    HEADER_ONLY = True
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("message_id", apt_word),
        ("chan_ident", apt_byte),
        ("enable_state", apt_byte),
        ("dest", apt_byte),
        ("source", apt_byte),
    ]


class MOT_MOVE_HOMED(AptMessage):
    """
    Header structure for the MOT_MOVE_HOMED response. This header is sent as a response to MOT_MOVE_HOME
    once homing is complete.

    Fields:
        message_id:     ID of message.
        chan_ident:     Channel number.
        param2:         To be left as 0x00.
        dest:           Destination of message.
        source:         Source of message.
    """

    MESSAGE_ID = AptMessageId.MOT_MOVE_HOMED.value
    HEADER_ONLY = True
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("message_id", apt_word),
        ("chan_ident", apt_byte),
        ("param2", apt_byte),
        ("dest", apt_byte),
        ("source", apt_byte),
    ]


class MOT_MOVE_ABSOLUTE(AptMessage):
    """
    Data packet structure for a MOT_SMOVE_ABSOLUTE command.

    Fields:
        chan_ident:         The channel being addressed.
        absolute_distance:  The distance to move in encoder units.
    """

    MESSAGE_ID = AptMessageId.MOT_MOVE_ABSOLUTE.value
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("chan_ident", apt_word),
        ("absolute_distance", apt_long),
    ]


class MOT_MOVE_COMPLETED(AptMessage):
    """
    Header structure for the MOT_MOVE_COMPLETED response. This header is sent as a response to a relative or absolute
    move command once the move has been completed.

    Fields:
        message_id:     ID of message.
        chan_ident:     Channel number.
        param2:         To be left as 0x00.
        dest:           Destination of message.
        source:         Source of message.
    """

    MESSAGE_ID = AptMessageId.MOT_MOVE_COMPLETED.value
    HEADER_ONLY = True
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("message_id", apt_word),
        ("chan_ident", apt_byte),
        ("param2", apt_byte),
        ("dest", apt_byte),
        ("source", apt_byte),
    ]


class MOT_GET_USTATUSUPDATE(AptMessage):
    """
    Data packet structure for a MOT_GET_USTATUSUPDATE command.

    Fields:
        chan_ident:     The channel being addressed.
        position:       The position in encoder counts.
        velocity:       Velocity in controller units. Note that this is reported as a 16 bit
                        unsigned integer in the manual but it is actually signed according to the example.
        motor_current:  Motor current in mA.
        status_bits:    Status bits that provide various errors and indications.
    """

    MESSAGE_ID = AptMessageId.MOT_GET_USTATUSUPDATE.value
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("chan_ident", apt_word),
        ("position", apt_long),
        ("velocity", apt_word),
        ("motor_current", apt_short),
        ("status_bits", apt_dword),
    ]


class MOT_SET_EEPROMPARAMS(AptMessage):
    """
    Data packet structure for a MOT_SET_EEPROMPARAMS command.

    Fields:
        chan_ident: The channel being addressed.
        msg_id:     ID of message whose settings should be save.
    """

    MESSAGE_ID = AptMessageId.MOT_SET_EEPROMPARAMS.value
    _fields_: ClassVar[list[tuple[str, type]]] = [("chan_ident", apt_word), ("msg_id", apt_word)]


class POL_GET_SET_PARAMS(AptMessage):
    """ "
    Data packet structure for POL_SET_PARAMS command. It is also the data packet structure for the POL_GET_PARAMS.

    Fields:
        not_used:       This field is not used, but needs to be in the field structure to not break it.
        velocity:       Velocity in range 10% to 100% of 400 degrees/s.
        home_position:  Home position in encoder counts.
        jog_step1:      Size fo jog step to be performed on paddle 1.
        jog_step2:      Size fo jog step to be performed on paddle 2.
        jog_step3:      Size fo jog step to be performed on paddle 3.
    """

    MESSAGE_ID = AptMessageId.POL_GET_PARAMS.value
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("not_used", apt_word),
        ("velocity", apt_word),
        ("home_position", apt_word),
        ("jog_step1", apt_word),
        ("jog_step2", apt_word),
        ("jog_step3", apt_word),
    ]



"""
Pico Technology Ltd., oscilloscopes.

The qmi.instruments.picotech package provides support for:
  - PicoScope 3403A and
  - PicoScope 4824A models.
"""
from qmi.instruments.picotech._picoscope import ChannelCoupling, TriggerEdge
from qmi.instruments.picotech.picoscope3403 import PicoTech_PicoScope3403
from qmi.instruments.picotech.picoscope4824 import PicoTech_PicoScope4824



import unittest
import os
import time

from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_ConfigurationException, QMI_UsageException


class TestQMIContext(unittest.TestCase):

    def setUp(self):
        # Start two contexts.
        self.log_dir = "~/log"
        self.datastore_dir = "~/datastore"
        self.config_file = "~/config/qmi.conf"
        self.qmi_home = "~/home"
        config1 = CfgQmi(
            contexts={"c1": CfgContext(tcp_server_port=0)},
            log_dir=self.log_dir,
            qmi_home=self.qmi_home,
            datastore=self.datastore_dir,
            config_file=self.config_file
            )
        config2 = CfgQmi(
            contexts={"c2": CfgContext(tcp_server_port=0)}
            )

        self.c1 = QMI_Context("c1", config1)
        self.c2 = QMI_Context("c2", config2)

        self.c1.start()
        self.c2.start()

    def tearDown(self):
        # Stop the contexts
        if self.c1._active:
            self.c1.stop()

        self.c1 = None
        if self.c2._active:
            self.c2.stop()

        self.c2 = None

    def test_version_mismatch_warnings(self):
        """Test that we can set and get version mismatch warnings"""
        suppress = self.c1.suppress_version_mismatch_warnings
        self.assertFalse(suppress)
        self.c1.suppress_version_mismatch_warnings = True
        suppress = self.c1.suppress_version_mismatch_warnings
        self.assertTrue(suppress)
        self.c1.suppress_version_mismatch_warnings = False

    def test_get_config(self):
        """Test that we get config object CfgQmi correctly."""
        # Arrange
        expected_config_file = self.config_file
        expected_log_dir = self.log_dir
        expected_home_dir = self.qmi_home
        expected_workgroup = "default"
        expected_context = {"c1": CfgContext(tcp_server_port=0)}
        expected_datastore = self.datastore_dir
        # Act
        config = self.c1.get_config()
        # Assert
        self.assertEqual(expected_config_file, config.config_file)
        self.assertEqual(expected_log_dir, config.log_dir)
        self.assertEqual(expected_home_dir, config.qmi_home)
        self.assertEqual(expected_workgroup, config.workgroup)
        self.assertEqual(expected_context, config.contexts)
        self.assertEqual(expected_datastore, config.datastore)

    def test_get_context_config(self):
        """Test that we get config context object CfgContext correctly."""
        # Arrange
        expected_context = CfgContext(tcp_server_port=0)
        # Act
        context = self.c2.get_context_config()
        # Assert
        self.assertEqual(expected_context, context)

    def test_get_qmi_home_dir(self):
        """Get the QMI home directory set in config and assert it is correct."""
        # Arrange
        expected_home = self.qmi_home
        # Act
        qmi_home = self.c1.get_qmi_home_dir()
        # Assert
        self.assertEqual(expected_home, qmi_home)

    def test_get_qmi_home_dir_default(self):
        """Get the QMI home directory reading it from environment variable 'QMI_HOME'."""
        # Arrange
        expected_home = os.path.expanduser("~")
        # Act
        qmi_home = self.c2.get_qmi_home_dir()
        # Assert
        self.assertEqual(expected_home, qmi_home)

    def test_get_log_dir(self):
        """Get the log directory set in config and assert it is correct."""
        # Arrange
        expected_log_dir = self.log_dir
        # Act
        log_dir = self.c1.get_log_dir()
        # Assert
        self.assertEqual(expected_log_dir, log_dir)

    def test_get_log_dir_default(self):
        """Get the log directory reading it from environment variable 'QMI_HOME'."""
        # Arrange
        expected_log_dir = os.path.expanduser("~")
        # Act
        log_dir = self.c2.get_log_dir()
        # Assert
        self.assertEqual(expected_log_dir, log_dir)

    def test_get_datastore_dir(self):
        """Try to get datastore directory from the config."""
        # Arrange
        expected_dir = self.datastore_dir
        # Act
        datastore_dir = self.c1.get_datastore_dir()
        # Assert
        self.assertEqual(expected_dir, datastore_dir)

    def test_get_datastore_dir_raises_exception(self):
        """Try to get datastore directory from the config, but it is not defined there, raising exception."""
        with self.assertRaises(QMI_ConfigurationException):
            self.c2.get_datastore_dir()

    def test_resolve_file_name(self):
        """Test that file name substitutions work for all configured file names. Names to be substituted are:
          $$            -> "$"
          ${context}    -> context name
          ${qmi_home}   -> QMI home directory
          ${datastore}  -> QMI datastore directory
          ${config_dir} -> directory of QMI configuration file
          ${date}       -> date of program start, UTC, formatted as YYYY-mm-dd
          ${datetime}   -> time of program start, UTC, formatted as YYYY-mm-ddTHH-MM-SS
        """
        # Test 1. File name not in substitutions list -> returns the input name as-is.
        expected_conf_file = "qmi.conf"
        resolved_name = self.c1.resolve_file_name("qmi.conf")
        self.assertEqual(expected_conf_file, resolved_name)
        # Test 2. Double dollar into one
        expected_name = "$"
        resolved_name = self.c1.resolve_file_name("$$")
        self.assertEqual(expected_name, resolved_name)
        # Test 3. Context name
        expected_context = "c1"
        resolved_name = self.c1.resolve_file_name("$context")
        self.assertEqual(expected_context, resolved_name)
        # Test 4. QMI home directory
        expected_directory = self.qmi_home
        resolved_name = self.c1.resolve_file_name("$qmi_home")
        self.assertEqual(expected_directory, resolved_name)
        # Test 5. Datastore directory
        expected_directory = self.datastore_dir
        resolved_name = self.c1.resolve_file_name("$datastore")
        self.assertEqual(expected_directory, resolved_name)
        # Test 6. Configuration file directory
        expected_directory = os.path.split(self.config_file)[0]
        resolved_name = self.c1.resolve_file_name("$config_dir")
        self.assertEqual(expected_directory, resolved_name)
        # Test 7. Date of program start
        gmtime = time.gmtime(self.c1._start_time)
        date = time.strftime("%Y-%m-%d", gmtime)
        resolved_name = self.c1.resolve_file_name("$date")
        self.assertEqual(date, resolved_name)
        # Test 8. Time of program start
        datetime = time.strftime("%Y-%m-%dT%H-%M-%S", gmtime)
        resolved_name = self.c1.resolve_file_name("$datetime")
        self.assertEqual(datetime, resolved_name)

    def test_resolve_without_datastore(self):
        """Test file name substitution when no datastore is configured."""

        resolved_name = self.c2.resolve_file_name("qmi.conf")
        self.assertEqual(resolved_name, "qmi.conf")

        resolved_name = self.c2.resolve_file_name("$context")
        self.assertEqual(resolved_name, "c2")

        with self.assertRaises(QMI_ConfigurationException):
            resolved_name = self.c2.resolve_file_name("$datastore")

    def test_resolve_datastore(self):
        """Test resolving file name substitution in the datastore path."""

        cfg = CfgQmi(
            contexts={"ctx_tmp": CfgContext(tcp_server_port=0)},
            log_dir=self.log_dir,
            qmi_home=self.qmi_home,
            datastore="${qmi_home}/datastore",
            config_file=self.config_file)

        ctx_tmp = QMI_Context("ctx_tmp", cfg)
        ctx_tmp.start()
        datastore_dir = ctx_tmp.get_datastore_dir()
        ctx_tmp.stop()
        self.assertEqual(datastore_dir, self.qmi_home + "/datastore")

    def test_resolve_recursive_datastore(self):
        """Test an attempt to recursively define the datastore path."""
        cfg = CfgQmi(
            contexts={"ctx_tmp": CfgContext(tcp_server_port=0)},
            log_dir=self.log_dir,
            qmi_home=self.qmi_home,
            datastore="${datastore}/datastore",
            config_file=self.config_file)

        ctx_tmp = QMI_Context("ctx_tmp", cfg)
        ctx_tmp.start()
        with self.assertRaises(QMI_ConfigurationException):
            datastore_dir = ctx_tmp.get_datastore_dir()
        ctx_tmp.stop()

    def test_list_rpc_objects(self):
        # Arrange
        expected_list_rpc_objects = [('c1.$context', '_ContextRpcObject'), ('c2.$context', '_ContextRpcObject')]
        # Act
        list_rpc_objects1 = self.c1.list_rpc_objects()
        list_rpc_objects2 = self.c2.list_rpc_objects()
        # Assert
        self.assertListEqual(expected_list_rpc_objects, list_rpc_objects1 + list_rpc_objects2)

    def test_shutdown_requested(self):
        requested = self.c1.shutdown_requested()
        self.assertFalse(requested)

    def test_wait_until_shutdown(self):
        shutdown = self.c1.wait_until_shutdown(0.01)
        self.assertFalse(shutdown)

    def test_double_start_excepts(self):
        with self.assertRaises(QMI_UsageException) as exc:
            self.c1.start()

        self.assertEqual("QMI_Context already started", str(exc.exception))

    def test_double_stop_excepts(self):
        self.c1.stop()
        with self.assertRaises(QMI_UsageException) as exc:
            self.c1.stop()

        self.assertEqual("QMI_Context already inactive", str(exc.exception))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock, call, patch

from string import punctuation

from qmi.core.transport import QMI_TcpTransport
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.instruments.pi import PI_E873, ReferenceTarget


class TestE8738(unittest.TestCase):
    def setUp(self):
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
            "qmi.instruments.pi.e873.create_transport",
            MagicMock(return_value=self._transport_mock),
        ):
            self.instr = PI_E873(MagicMock(), "instr", "transport_descriptor")

    def tearDown(self):
        self.instr = None

    def _instr_open(self):
        """Open the instrument and clear the write mocked calls."""
        self._transport_mock.read_until.return_value = b"1\r\n"
        self.instr.open()
        self._transport_mock.write.reset_mock()

    def _test_parse_invalid(self, function, response):
        with self.assertRaises(QMI_InstrumentException):
            function(response, "", "")

    def _test_get_macros(self, function, command):
        self._instr_open()
        self._transport_mock.read_until.side_effect = [
            b"test1 \n",
            b"test2 \n",
            b"test3\n",
        ]
        m = function()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(
            f"{command}\n".encode("UTF-8")
        )
        self.assertEqual(m, ["test1", "test2", "test3"])

    def _test_variable_name_invalid(self, function, *args, **kwargs):
        def set_variable_helper(name):
            self._instr_open()
            with self.assertRaises(ValueError):
                function(name, *args, **kwargs)
            self.instr.close()

        set_variable_helper("A" * 9)  # char limit
        set_variable_helper("1")  # should start with a letter
        set_variable_helper(" BLA")  # should start with a letter
        set_variable_helper("BLA ")  # should end with a letter
        for i in punctuation:  # special characters
            set_variable_helper("BLA" + i)

    def _test_set_cmd_function(self, function, command):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        function(0.001)
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [call(command.encode("UTF-8") + b" 1 0.00100000\n"), call(b"ERR?\n")]
        )

    def _test_float_get_address_function(self, function, address):
        self._instr_open()
        t = f"1 0X{address:X}".encode("UTF-8")
        self._transport_mock.read_until.return_value = t + b"=0.001\n"
        r = function()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"SPA? " + t + b"\n")
        self.assertEqual(r, 0.001)

    def _test_float_get_cmd_function(self, function, command):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"1=0.001\n"
        r = function()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(
            f"{command} 1\n".encode("UTF-8")
        )
        self.assertEqual(r, 0.001)

    def _test_bool_set_cmd_function(self, function, command):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        function(1)
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [call(command.encode("UTF-8") + b" 1 1\n"), call(b"ERR?\n")]
        )

    def _test_bool_get_cmd_function(self, function, command):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"1=1\n"
        r = function()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(
            f"{command} 1\n".encode("UTF-8")
        )
        self.assertEqual(r, 1)

    def _test_dig_out_invalid(self, function, *args, **kwargs):
        self._instr_open()
        with self.assertRaises(QMI_InstrumentException):
            function(5, *args, **kwargs)
        self.instr.close()

    def _test_set_digital_output(self, function, command):
        for i in [1, 2, 3, 4]:
            self._instr_open()
            self._transport_mock.read_until.return_value = b"0\n"
            function(i, True)
            self._transport_mock.write.assert_has_calls(
                [
                    call(f"{command} {i} 1\n".encode("UTF-8")),
                    call(b"ERR?\n"),
                ]
            )
            self.instr.close()

    def _test_get_digital_output(self, function, command):
        for i in [1, 2, 3, 4]:
            self._instr_open()
            self._transport_mock.read_until.return_value = f"{i}=1\n".encode("UTF-8")
            s = function(i)
            self.instr.close()
            self._transport_mock.write.assert_called_once_with(
                f"{command} {i}\n".encode("UTF-8")
            )
            self.assertEqual(s, 1)

    def test_open_close(self):
        self._transport_mock.read_until.return_value = b"1\r\n"
        self.instr.open()
        self.instr.close()

        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.write.assert_has_calls([call(b"ERR?\n"), call(b"SAI?\n")])
        self._transport_mock.close.assert_called_once_with()

    def test_open_invalid(self):
        self._transport_mock.read_until.return_value = b"2\r\n"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.open()
        self.instr.close()

    def test_ask_invalid(self):
        # Build a mock instrument to access ._ask
        instr = MagicMock(spec=PI_E873)
        instr._is_open = MagicMock(return_value=True)
        instr._transport = self._transport_mock

        def invalid_helper(value, exception):
            self._transport_mock.read_until.return_value = f"{value}".encode("UTF-8")
            with self.assertRaises(exception):
                PI_E873._ask(instr, "CMD")

        invalid_helper("", QMI_TimeoutException)
        invalid_helper(" ", QMI_InstrumentException)

    def test_parse_response_item_invalid(self):
        self._test_parse_invalid(PI_E873._parse_response_item, "\n\n")
        self._test_parse_invalid(PI_E873._parse_response_item, "NOKEY\n")

    def test_parse_int_invalid(self):
        self._test_parse_invalid(PI_E873._parse_int, "NOTINT")

    def test_parse_float_invalid(self):
        self._test_parse_invalid(PI_E873._parse_float, "NOTFLOAT")

    def test_reset(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.reset()
        self.instr.close()

        self._transport_mock.write.assert_has_calls([call(b"RBT\n"), call(b"ERR?\n")])

    def test_get_idn(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = (
            b"vendor,model,serial,version\r\n"
        )
        i = self.instr.get_idn()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"*IDN?\n")
        self.assertEqual(i.vendor, "vendor")
        self.assertEqual(i.model, "model")
        self.assertEqual(i.serial, "serial")
        self.assertEqual(i.version, "version")

    def test_get_idn_invalid(self):
        self._instr_open()
        with self.assertRaises(QMI_InstrumentException):
            self._transport_mock.read_until.return_value = b"\n"
            self.instr.get_idn()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"*IDN?\n")

    def test_get_stage_info(self):
        self._instr_open()
        self._transport_mock.read_until.side_effect = r = []
        c = []
        for i in [1, 2, 3, 4]:
            t = f"1 0X{0X0F000000 + (i << 8):X}"
            r.append(f"{t}=result{i}\n".encode("UTF-8"))
            c.append(call(f"SPA? {t}\n".encode("UTF-8")))
        s = self.instr.get_stage_info()
        self.instr.close()

        self._transport_mock.write.assert_has_calls(c)
        self.assertEqual(s.type, "result1")
        self.assertEqual(s.serial_number, "result2")
        self.assertEqual(s.assembly_date, "result3")
        self.assertEqual(s.hw_version, "result4")

    def test_get_error(self):
        def get_error_helper(value):
            self._instr_open()
            self._transport_mock.read_until.return_value = f"{value}\n".encode("UTF-8")
            e = self.instr.get_error()
            self.instr.close()

            self._transport_mock.write.assert_called_once_with(b"ERR?\n")
            self.assertEqual(e, value)

        get_error_helper(0)
        get_error_helper(1)

    def test_get_system_status(self):
        sm = {
            # id: (shift, mask, val)
            "negative_limit_switch": (0, 0x1, 1),
            "reference_point_switch": (1, 0x1, 1),
            "positive_limit_switch": (2, 0x1, 1),
            "digital_input": (4, 0xF, (1, 1, 1, 1)),
            "error_flag": (8, 0x1, 1),
            "servo_mode": (12, 0x1, 1),
            "in_motion": (13, 0x1, 1),
            "on_target": (15, 0x1, 1),
        }

        for key, (s, m, v) in sm.items():
            self._instr_open()
            self._transport_mock.read_until.return_value = f"{m << s}\n".encode("UTF-8")
            ss = self.instr.get_system_status()
            self.instr.close()

            self._transport_mock.write.assert_called_once_with(b"\x04")
            self.assertEqual(
                i := getattr(ss, key), v, msg=f"{key}: is {i} expected {v}."
            )

    def test_stop_all(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"10\n"
        self.instr.stop_all()

        self._transport_mock.write.assert_any_call(b"\x18")
        self._transport_mock.write.assert_called_with(b"ERR?\n")

        self.instr.close()

    def test_stop_all_with_error(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"1\n"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.stop_all()
        self.instr.close()

    def test_stop_smooth(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"10\n"
        self.instr.stop_smooth()
        self.instr.close()

        self._transport_mock.write.assert_any_call(b"HLT\n")
        self._transport_mock.write.assert_called_with(b"ERR?\n")

    def test_get_physical_unit(self):
        self._instr_open()
        t = f"1 0X{0x07000601:X}".encode("UTF-8")
        self._transport_mock.read_until.return_value = t + b"=result\n"
        p = self.instr.get_physical_unit()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"SPA? " + t + b"\n")
        self.assertEqual(p, "result")

    def test_get_position_range(self):
        self._instr_open()
        self._transport_mock.read_until.side_effect = [
            b"1=0.001\r\n",
            b"1=0.002\r\n",
        ]
        p = self.instr.get_position_range()
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [call(b"TMN? 1\n"), call(b"TMX? 1\n")]
        )
        self.assertEqual((0.001, 0.002), p)

    def test_get_max_position_error(self):
        self._test_float_get_address_function(self.instr.get_max_position_error, 0x8)

    def test_get_max_velocity(self):
        self._test_float_get_address_function(self.instr.get_max_velocity, 0xA)

    def test_get_max_acceleration(self):
        self._test_float_get_address_function(self.instr.get_max_acceleration, 0x4A)

    def test_get_max_deceleration(self):
        self._test_float_get_address_function(self.instr.get_max_deceleration, 0x4B)

    def test_get_reference_signal_mode(self):
        self._instr_open()
        t = f"1 0X{0x70:X}".encode("UTF-8")
        self._transport_mock.read_until.return_value = t + b"=1\n"
        m = self.instr.get_reference_signal_mode()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"SPA? " + t + b"\n")
        self.assertEqual(m.value, 1)

    def test_get_reference_velocity(self):
        self._test_float_get_address_function(self.instr.get_reference_velocity, 0x50)

    def test_set_acceleration(self):
        self._test_set_cmd_function(self.instr.set_acceleration, "ACC")

    def test_get_acceleration(self):
        self._test_float_get_cmd_function(self.instr.get_acceleration, "ACC?")

    def test_set_deceleration(self):
        self._test_set_cmd_function(self.instr.set_deceleration, "DEC")

    def test_get_deceleration(self):
        self._test_float_get_cmd_function(self.instr.get_deceleration, "DEC?")

    def test_set_velocity(self):
        self._test_set_cmd_function(self.instr.set_velocity, "VEL")

    def test_set_velocity_negative(self):
        self._instr_open()
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_velocity(-1)
        self.instr.close()

    def test_get_velocity(self):
        self._test_float_get_cmd_function(self.instr.get_velocity, "VEL?")

    def test_set_servo_mode(self):
        self._test_bool_set_cmd_function(self.instr.set_servo_mode, "SVO")

    def test_get_servo_mode(self):
        self._test_bool_get_cmd_function(self.instr.get_servo_mode, "SVO?")

    def test_set_reference_definition_mode(self):
        self._test_bool_set_cmd_function(
            self.instr.set_reference_definition_mode, "RON"
        )

    def test_get_reference_definition_mode(self):
        self._test_bool_get_cmd_function(
            self.instr.get_reference_definition_mode, "RON?"
        )

    def test_reference_move(self):
        def reference_move_helper(target, command):
            self._instr_open()
            self._transport_mock.read_until.return_value = b"0\n"
            self.instr.reference_move(target)
            self.instr.close()

            self._transport_mock.write.assert_has_calls(
                [call(command.encode("UTF-8") + b"\n"), call(b"ERR?\n")]
            )

        reference_move_helper(ReferenceTarget.REFERENCE_POINT, "FRF")
        reference_move_helper(ReferenceTarget.POSITIVE_LIMIT, "FPL")
        reference_move_helper(ReferenceTarget.NEGATIVE_LIMIT, "FNL")

    def test_get_reference_result(self):
        self._test_bool_get_cmd_function(self.instr.get_reference_result, "FRF?")

    def test_move_absolute(self):
        self._test_set_cmd_function(self.instr.move_absolute, "MOV")

    def test_move_relative(self):
        self._test_set_cmd_function(self.instr.move_relative, "MVR")

    def test_get_target_position(self):
        self._test_float_get_cmd_function(self.instr.get_target_position, "MOV?")

    def test_get_position(self):
        self._test_float_get_cmd_function(self.instr.get_position, "POS?")

    def test_wait_motion_complete(self):
        self._instr_open()
        self._transport_mock.read_until.side_effect = [
            f"{1<<13}\r\n".encode("UTF-8"),
            f"{1<<13}\r\n".encode("UTF-8"),
            f"{0<<13}\r\n".encode("UTF-8"),
        ]
        r = self.instr.wait_motion_complete()
        self.instr.close()
        self.assertEqual(r, True)

    def test_wait_motion_complete_timeout(self):
        self._instr_open()
        self._transport_mock.read_until.side_effect = [
            f"{1<<13}\r\n".encode("UTF-8"),
            f"{1<<13}\r\n".encode("UTF-8"),
            f"{1<<13}\r\n".encode("UTF-8"),
            f"{0<<13}\r\n".encode("UTF-8"),
        ]

        with patch("qmi.instruments.pi.e873.time.monotonic", side_effect=[0, 0, 100]):
            r = self.instr.wait_motion_complete(100)
            self.instr.close()

        self.assertEqual(r, False)

    def test_set_trigger_inmotion(self):
        for i in [1, 2, 3, 4]:
            self._instr_open()
            self._transport_mock.read_until.return_value = b"0\n"
            self.instr.set_trigger_inmotion(i)
            self.instr.close()
            self._transport_mock.write.assert_has_calls(
                [
                    call(f"CTO {i} 2 1\n".encode("UTF-8")),
                    call(f"CTO {i} 3 6\n".encode("UTF-8")),
                    call(b"ERR?\n"),
                ]
            )

    def test_set_trigger_inmotion_invalid(self):
        self._test_dig_out_invalid(self.instr.set_trigger_inmotion)

    def test_set_trigger_position_offset(self):
        for i in [1, 2, 3, 4]:
            self._instr_open()
            self._transport_mock.read_until.return_value = b"0\n"
            self.instr.set_trigger_position_offset(i, 0.1, 0.2, 0.3)
            self.instr.close()
            self._transport_mock.write.assert_has_calls(
                [
                    call(f"CTO {i} 2 1\n".encode("UTF-8")),
                    call(f"CTO {i} 1 {0.1:.8f}\n".encode("UTF-8")),
                    call(f"CTO {i} 10 {0.2:.8f}\n".encode("UTF-8")),
                    call(f"CTO {i} 9 {0.3:.8f}\n".encode("UTF-8")),
                    call(f"CTO {i} 3 7\n".encode("UTF-8")),
                    call(b"ERR?\n"),
                ]
            )

    def test_set_trigger_position_offset_invalid(self):
        self._test_dig_out_invalid(
            self.instr.set_trigger_position_offset, 0.1, 0.2, 0.3
        )

    def test_set_trigger_output_state(self):
        self._test_set_digital_output(self.instr.set_trigger_output_state, "TRO")

    def test_set_trigger_output_state_invalid(self):
        self._test_dig_out_invalid(self.instr.set_trigger_output_state, True)

    def test_get_trigger_output_state(self):
        self._test_get_digital_output(self.instr.get_trigger_output_state, "TRO?")

    def test_get_trigger_output_state_invalid(self):
        self._test_dig_out_invalid(self.instr.get_trigger_output_state)

    def test_set_digital_output(self):
        self._test_set_digital_output(self.instr.set_digital_output, "DIO")

    def test_set_digital_output_invalid(self):
        self._test_dig_out_invalid(self.instr.set_digital_output, True)

    def test_get_digital_input(self):
        self._test_get_digital_output(self.instr.get_digital_input, "DIO?")

    def test_get_digital_input_invalid(self):
        self._test_dig_out_invalid(self.instr.get_digital_input)

    def test_define_macro(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        c = ["TESTCMD1", "TESTCMD2", "TESTCMD3"]
        self.instr.define_macro("test", c)
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"MAC BEG test\n"),
                call(b"TESTCMD1\n"),
                call(b"TESTCMD2\n"),
                call(b"TESTCMD3\n"),
                call(b"MAC END\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_delete_macro(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.delete_macro("test")
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"MAC DEL test\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_get_defined_macros(self):
        self._test_get_macros(self.instr.get_defined_macros, "MAC?")

    def test_start_macro(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.start_macro("test", "arg1", "arg2")
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"MAC START test arg1 arg2\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_start_macro_repeat(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.start_macro("test", "arg1", "arg2", 3)
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"MAC NSTART test 3 arg1 arg2\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_start_macro_invalid(self):
        self._instr_open()
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_macro("test", None, "arg2", 3)
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_macro("test", "arg1", "arg2", 0)
        self.instr.close()

    def test_get_running_macros(self):
        self._test_get_macros(self.instr.get_running_macros, "RMC?")

    def test_set_variable(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.set_variable("test", None)
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"VAR test\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_set_variable_value(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"0\n"
        self.instr.set_variable("test", "valuetest")
        self.instr.close()

        self._transport_mock.write.assert_has_calls(
            [
                call(b"VAR test valuetest\n"),
                call(b"ERR?\n"),
            ]
        )

    def test_set_variable_invalid(self):
        self._test_variable_name_invalid(self.instr.set_variable, "")

    def test_get_variable(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"test=test1\n"
        variable = self.instr.get_variable("test")
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"VAR? test\n")
        self.assertEqual(variable, "test1")

    def test_get_variable_invalid(self):
        self._test_variable_name_invalid(self.instr.get_variable)

    def test_get_variables(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = (
            b"test1=val1\ntest2=val2\ntest3=val3\n"
        )

        v = self.instr.get_variables()
        self.instr.close()

        self._transport_mock.write.assert_called_once_with(b"VAR?\n")
        self.assertEqual(
            v,
            {
                "test1": "val1",
                "test2": "val2",
                "test3": "val3",
            },
        )

    def test_get_variables_invalid(self):
        self._instr_open()
        self._transport_mock.read_until.return_value = b"NOKEY\n"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_variables()
        self.instr.close()


if __name__ == "__main__":
    unittest.main()



"""Test for the Wavelength TC Lab driver."""

import logging

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.instruments.wavelength import Wavelength_TcLab
from qmi.instruments.wavelength import AutotuneMode
from qmi.core.exceptions import QMI_InstrumentException


# Disable all logging
logging.disable(logging.CRITICAL)


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    scpi: MagicMock
    super: MagicMock
    instr: Wavelength_TcLab


class TestParameters:
    """Test parameter data for UsbTmcTransportDescriptorParser."""

    def __init__(self, vendor_id, product_id):
        self.vendorid = vendor_id
        self.productid = product_id

    def get(self, str):
        return getattr(self, str)


class TestWavelengthTCLab(TestCase):
    """Testcase for the Wavelength_TcLab class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_scpi = MagicMock(spec=ScpiProtocol)
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.wavelength.tclab.create_transport",
            return_value=mock_transport,
        ), patch("qmi.instruments.wavelength.tclab.ScpiProtocol", mock_scpi):
            instr = Wavelength_TcLab(MagicMock(), mock_name, mock_transport_str)

        self._patcher_super = patch(
            "qmi.instruments.wavelength.tclab.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(Wavelength_TcLab, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            scpi=mock_scpi(mock_transport),
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_list_instruments(self):
        """Wavelength_TcLab.list_instruments(), happy flow."""
        map = [
            {"instr": MagicMock(), "param": TestParameters(0x1A45, 0x3101)},
            {"instr": MagicMock(), "param": TestParameters(0x0000, 0x3101)},
            {"instr": MagicMock(), "param": TestParameters(0x1A45, 0x0000)},
            {"instr": MagicMock(), "param": TestParameters(0x1A45, 0x3101)},
        ]
        mock_lister = MagicMock(return_value=[item["instr"] for item in map])

        def helper_parse(transport):
            for item in map:
                if item["instr"] == transport:
                    return item["param"]

        mock_parser = MagicMock()
        mock_parser.parse_parameter_strings = MagicMock(side_effect=helper_parse)

        with patch(
            "qmi.instruments.wavelength.tclab.list_usbtmc_transports", mock_lister
        ), patch(
            "qmi.instruments.wavelength.tclab.UsbTmcTransportDescriptorParser",
            mock_parser,
        ):
            rt_val = self._meta.instr.list_instruments()

        self.assertTrue(map[0]["instr"] in rt_val)
        self.assertTrue(map[3]["instr"] in rt_val)

    def test_init(self):
        """Wavelength_TcLab.__init__(), happy flow."""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)
        self.assertEqual(self._meta.instr._scpi_protocol, self._meta.scpi)

    def test_open_transport(self):
        """Wavelength_TcLab._open_transport(), happy flow."""
        self._meta.instr._open_transport()
        self._meta.transport.open.assert_called_once_with()

    def test_open_transport_retry_exceed(self):
        """Wavelength_TcLab._open_transport(), retry exceeded."""
        mock_open_max_retry = 3
        self._meta.instr.OPEN_MAX_RETRY = mock_open_max_retry
        self._meta.transport.open = MagicMock(side_effect=Exception)
        with patch("qmi.instruments.wavelength.tclab.time", MagicMock()):
            with self.assertRaises(Exception):
                self._meta.instr._open_transport()
        self._meta.transport.open.assert_has_calls([call()] * mock_open_max_retry)

    def test_open(self):
        """Wavelength_TcLab.open(), happy flow."""
        self._meta.instr._check_is_closed = MagicMock()
        self._meta.instr._open_transport = MagicMock()

        self._meta.instr.open()
        self._meta.scpi.write.assert_has_calls([call("*CLS"), call("RADIX DEC")])
        self._meta.scpi.ask.assert_called_once_with("ERRSTR?")
        self._meta.super().open.assert_called_once_with()
        self._meta.instr._check_is_closed.assert_called_once_with()
        self._meta.instr._open_transport.assert_called_once_with()

    def test_open_exception(self):
        """Wavelength_TcLab.open(), exception handling."""
        self._meta.instr._check_is_closed = MagicMock()
        self._meta.instr._open_transport = MagicMock()
        self._meta.scpi.write = MagicMock(side_effect=Exception)

        with self.assertRaises(Exception):
            self._meta.instr.open()

        self._meta.transport.close.assert_called_once_with()
        self._meta.instr._check_is_closed.assert_called_once_with()
        self._meta.instr._open_transport.assert_called_once_with()

    def test_close(self):
        """Wavelength_TcLab.close(), happy flow."""
        self._meta.instr.close()
        self._meta.super().close.assert_called_once_with()
        self._meta.transport.close.assert_called_once_with()

    def test_check_error(self):
        """Wavelength_TcLab._check_error(), happy flow."""
        mock_resp = "0,No error"
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        self._meta.instr._check_error()
        self._meta.scpi.ask.assert_called_once_with("ERRSTR?")

    def test_check_error_is_error(self):
        """Wavelength_TcLab._check_error(), is error."""
        mock_resp = "1,No error"
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._check_error()
        self._meta.scpi.ask.assert_called_once_with("ERRSTR?")

    def test_ask_int(self):
        """Wavelength_TcLab._ask_int(), happy flow."""
        mock_resp = 18
        mock_cmd = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._ask_int(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self.assertEqual(rt_val, mock_resp)

    def test_ask_int_exception(self):
        """Wavelength_TcLab._ask_int(), exception handling."""
        mock_resp = "notanint"
        mock_cmd = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._ask_int(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)

    def test_ask_float(self):
        """Wavelength_TcLab.ask_float(), happy flow."""
        mock_resp = 0.001
        mock_cmd = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._ask_float(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self.assertEqual(rt_val, mock_resp)

    def test_ask_float_exception(self):
        """Wavelength_TcLab._ask_float(), exception handling."""
        mock_resp = "notafloat"
        mock_cmd = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._ask_float(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)

    def test_reset(self):
        """Wavelength_TcLab.reset(), happy flow."""
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.reset()
        self._meta.scpi.write.assert_has_calls([call("*CLS"), call("*RST")])
        self._meta.scpi.ask.assert_called_once_with("*OPC?")
        self._meta.instr._check_error.assert_called_once_with()

    def test_get_idn(self):
        """Wavelength_TcLab.get_idn(), happy flow."""
        mock_resp = "VENDOR,MODEL,SERIAL,VERSION"
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_idn()
        self._meta.scpi.ask.assert_called_once_with("*IDN?")
        self.assertEqual(rt_val.vendor, "VENDOR")
        self.assertEqual(rt_val.model, "MODEL")
        self.assertEqual(rt_val.serial, "SERIAL")
        self.assertEqual(rt_val.version, "VERSION")

    def test_get_idn_invalid_response(self):
        """Wavelength_TcLab.get_idn(), exception handling."""
        mock_resp = ""
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_idn()

    def test_get_power_on(self):
        """Wavelength_TcLab.get_power_on(), happy flow."""
        mock_resp = MagicMock()
        self._meta.instr._ask_int = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_power_on()
        self._meta.instr._ask_int.assert_called_once_with("PWR?")
        self.assertEqual(rt_val, bool(mock_resp))

    def _test_set_bool(self, function_name, expected_str):
        mock_power_on = [True, False]
        for item in mock_power_on:
            self._meta.scpi.write = MagicMock()
            self._meta.instr._check_error = MagicMock()
            getattr(self._meta.instr, function_name)(item)
            self._meta.scpi.write.assert_called_once_with(f"{expected_str} {int(item)}")
            self._meta.instr._check_error.assert_called_once_with()

    def test_set_power_on(self):
        """Wavelength_TcLab.set_power_on(), happy flow."""
        self._test_set_bool("set_power_on", "PWR")

    def test_set_output_enabled(self):
        """Wavelength_TcLab.set_output_enabled(), happy flow."""
        self._test_set_bool("set_output_enabled", "TEC:OUTPUT")

    def _test_get_float(self, function_name, expected_string):
        mock_resp = MagicMock()
        self._meta.instr._ask_float = MagicMock(return_value=mock_resp)
        rt_val = getattr(self._meta.instr, function_name)()
        self._meta.instr._ask_float.assert_called_once_with(expected_string)
        self.assertEqual(rt_val, mock_resp)

    def test_get_temperature(self):
        """Wavelength_TcLab.get_temperature(), happy flow."""
        self._test_get_float("get_temperature", "TEC:ACT?")

    def test_get_setpoint(self):
        """Wavelength_TcLab.get_setpoint(), happy flow."""
        self._test_get_float("get_setpoint", "TEC:SET?")

    def test_get_tec_current(self):
        """Wavelength_TcLab.get_tec_current(), happy flow."""
        self._test_get_float("get_tec_current", "TEC:I?")

    def test_get_tec_voltage(self):
        """Wavelength_TcLab.get_tec_voltage(), happy flow."""
        self._test_get_float("get_tec_voltage", "TEC:V?")

    def test_get_tec_voltage_limit(self):
        """Wavelength_TcLab.get_tec_voltage_limit(), happy flow."""
        self._test_get_float("get_tec_voltage_limit", "TEC:VLIM?")

    def test_set_setpoint(self):
        """Wavelength_TcLab.set_setpoint(), happy flow."""
        mock_setpoint = 0.001
        self._meta.scpi.write = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.set_setpoint(mock_setpoint)
        self._meta.scpi.write.assert_called_once_with(
            "TEC:SET {:.6f}".format(mock_setpoint)
        )
        self._meta.instr._check_error.assert_called_once_with()

    def test_get_unit(self):
        """Wavelength_TcLab.get_unit(), happy flow."""
        mock_resp = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_unit()
        self._meta.scpi.ask.assert_called_once_with("TEC:UNITS?")
        self.assertEqual(rt_val, mock_resp)

    def test_set_unit(self):
        """Wavelength_TcLab.set_unit(), happy flow."""
        mock_unit = "celsius"
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.set_unit(mock_unit)
        self._meta.scpi.write.assert_called_once_with("TEC:UNITS CELSIUS")
        self._meta.instr._check_error.assert_called_once_with()

    def test_set_unit_value_error(self):
        """Wavelength_TcLab.set_unit(), exception handling."""
        mock_unit = "INVALID"
        with self.assertRaises(ValueError):
            self._meta.instr.set_unit(mock_unit)

    def test_get_output_enabled(self):
        """Wavelength_TcLab.get_output_enabled(), happy flow."""
        mock_resp = MagicMock()
        self._meta.instr._ask_int = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_output_enabled()
        self._meta.instr._ask_int.assert_called_once_with("TEC:OUTPUT?")
        self.assertEqual(rt_val, bool(mock_resp))

    def test_get_condition_status(self):
        """Wavelength_TcLab.get_condition_status(), happy flow."""
        map = [
            {"name": "current_limit", "mask": 1},
            {"name": "sensor_limit", "mask": 4},
            {"name": "temperature_high", "mask": 8},
            {"name": "temperature_low", "mask": 16},
            {"name": "sensor_shorted", "mask": 32},
            {"name": "sensor_open", "mask": 64},
            {"name": "tec_open", "mask": 128},
            {"name": "in_tolerance", "mask": 512},
            {"name": "output_on", "mask": 1024},
            {"name": "laser_shutdown", "mask": 2048},
            {"name": "power_on", "mask": 32768},
        ]
        for item in map:
            self._meta.instr._ask_int = MagicMock(return_value=item["mask"])
            rt_val = self._meta.instr.get_condition_status()
            for item2 in map:
                val = getattr(rt_val, item2["name"])
                if item["name"] == item2["name"]:
                    self.assertEqual(
                        val,
                        True,
                        msg=f"{item['name']}, {item2['name']}, {item['mask']}, {rt_val}",
                    )
                else:
                    self.assertEqual(
                        val,
                        False,
                        msg=f"{item['name']}, {item2['name']}, {item['mask']}, {rt_val}",
                    )
            self._meta.instr._ask_int.assert_called_once_with("TEC:COND?")

    def test_get_pid_parameter(self):
        """Wavelength_TcLab.get_pid_parameter(), happy flow."""
        mock_resp = "0.01,0.02,0.03"
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val1, rt_val2, rt_val3 = self._meta.instr.get_pid_parameters()
        self._meta.scpi.ask.assert_called_once_with("TEC:PID?")
        self.assertEqual(rt_val1, 0.01)
        self.assertEqual(rt_val2, 0.02)
        self.assertEqual(rt_val3, 0.03)

    def test_get_pid_parameter_invalid_response(self):
        """Wavelength_TcLab.get_pid_parameter(), invalid response."""
        mock_resp = ""
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_pid_parameters()
        self._meta.scpi.ask.assert_called_once_with("TEC:PID?")

    def test_get_pid_parameter_invalid_value(self):
        """Wavelength_TcLab.get_pid_parameter(), invalid value."""
        mock_resp = "notvalid,notvalid,notvalid"
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_pid_parameters()
        self._meta.scpi.ask.assert_called_once_with("TEC:PID?")

    def test_set_pid_parameter(self):
        """Wavelength_TcLab.set_pid_parameter(), happy flow."""
        mock_p = 0.001
        mock_i = 0.002
        mock_d = 0.003
        self._meta.scpi.write = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.set_pid_parameters(mock_p, mock_i, mock_d)
        self._meta.scpi.write.assert_called_once_with("TEC:PID 0.0010,0.0020,0.0030")
        self._meta.instr._check_error.assert_called_once_with()

    def test_get_autotune_mode(self):
        """Wavelength_TcLab.get_autotune_mode(), happy flow."""
        map = [
            {"resp": 0, "expected_rt_val": AutotuneMode.MANUAL},
            {"resp": 1, "expected_rt_val": AutotuneMode.DISTURB_REJECT},
            {"resp": 2, "expected_rt_val": AutotuneMode.SETPOINT_RESPONSE},
        ]

        for item in map:
            self._meta.instr._ask_int = MagicMock(return_value=item["resp"])
            rt_val = self._meta.instr.get_autotune_mode()
            self.assertEqual(rt_val, item["expected_rt_val"])
            self._meta.instr._ask_int.assert_called_once_with("TEC:AUTOTUNE?")

    def test_set_autotune_mode(self):
        """Wavelength_TcLab.set_autotune_mode(), happy flow."""
        mock_mode = AutotuneMode.DISTURB_REJECT
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.set_autotune_mode(mock_mode)
        self._meta.scpi.write.assert_called_once_with(f"TEC:AUTOTUNE {mock_mode.value}")
        self._meta.instr._check_error.assert_called_once_with()

    def test_start_autotune(self):
        """Wavelength_TcLab.start_autotune(), happy flow."""
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.start_autotune()
        self._meta.scpi.write.assert_called_once_with("TEC:TUNESTART")
        self._meta.instr._check_error.assert_called_once_with()

    def test_abort_autotune(self):
        """Wavelength_TcLab.abort_autotune(), happy flow."""
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.abort_autotune()
        self._meta.scpi.write.assert_called_once_with("TEC:TUNEABORT")
        self._meta.instr._check_error.assert_called_once_with()

    def test_get_autotune_is_valid(self):
        """Wavelength_TcLab.get_autotune_is_valid(), happy flow."""
        mock_resp = MagicMock()
        self._meta.instr._ask_int = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_autotune_is_valid()
        self.assertEqual(rt_val, bool(mock_resp))
        self._meta.instr._ask_int.assert_called_once_with("TEC:VALID?")

    def _test_get_float_tuple(self, function_name, test_cmd_1, test_cmd_2):
        mock_resp_1 = MagicMock()
        mock_resp_2 = MagicMock()

        def se_ask_float(cmd):
            if cmd == test_cmd_1:
                return mock_resp_1
            elif cmd == test_cmd_2:
                return mock_resp_2

        self._meta.instr._ask_float = MagicMock(side_effect=se_ask_float)
        rt_val = getattr(self._meta.instr, function_name)()
        self.assertEqual(rt_val[0], mock_resp_1)
        self.assertEqual(rt_val[1], mock_resp_2)

    def test_get_temperature_limit(self):
        """Wavelength_TcLab.get_temperature_limit(), happy flow."""
        self._test_get_float_tuple(
            "get_temperature_limit", "TEC:LIM:TLO?", "TEC:LIM:THI?"
        )

    def test_get_tec_current_limit(self):
        """Wavelength_TcLab.get_tec_current_limit(), happy flow."""
        self._test_get_float_tuple(
            "get_tec_current_limit", "TEC:LIM:IPOS?", "TEC:LIM:INEG?"
        )

    def _test_set_scpi_tuple(self, function_name, test_cmd_1, test_cmd_2):
        mock_arg_1 = 0.001
        mock_arg_2 = 0.002

        self._meta.instr._check_error = MagicMock()
        self._meta.scpi.write = MagicMock()
        getattr(self._meta.instr, function_name)(mock_arg_1, mock_arg_2)
        self._meta.scpi.write.assert_has_calls(
            [
                call(f"{test_cmd_1} {mock_arg_1:.4f}"),
                call(f"{test_cmd_2} {mock_arg_2:.4f}"),
            ]
        )
        self._meta.instr._check_error.assert_called_once_with()

    def test_set_temperature_limit(self):
        """Wavelength_TcLab.set_temperature_limit(), happy flow."""
        self._test_set_scpi_tuple("set_temperature_limit", "TEC:LIM:TLO", "TEC:LIM:THI")

    def test_set_tec_current_limit(self):
        """Wavelength_TcLab.set_tec_current_limit(), happy flow."""
        self._test_set_scpi_tuple(
            "set_tec_current_limit", "TEC:LIM:IPOS", "TEC:LIM:INEG"
        )

    def test_set_tec_voltage_limit(self):
        """Wavelength_TcLab.set_tec_voltage_limit(), happy flow."""
        mock_vlim = 0.001
        self._meta.instr._check_error = MagicMock()
        self._meta.scpi.write = MagicMock()
        self._meta.instr.set_tec_voltage_limit(mock_vlim)
        self._meta.scpi.write.assert_called_once_with(f"TEC:VLIM {mock_vlim:.4f}")
        self._meta.instr._check_error.assert_called_once_with()



"""
Exception definitions for the QMI framework.
"""


class QMI_Exception(Exception):
    """Base class for QMI exceptions."""


class QMI_RuntimeException(QMI_Exception):
    """Raised when an error occurs that does not fit in any other category."""


class QMI_ApplicationException(QMI_Exception):
    """Raised in top-level application code to exit with an error message."""


#
# Exceptions typically caused by incorrect use of QMI:
#

class QMI_UsageException(QMI_Exception):
    """Raised when the library is used in a way that was not intended.

    Examples:
     - Bad data type of argument
     - Invalid order of operations
     - Use without initialization
    """


class QMI_DuplicateNameException(QMI_Exception):
    """Raised when a duplicate name is assigned while a unique name is required."""


class QMI_UnknownNameException(QMI_Exception):
    """Raised when an unknown name is used where an assigned name is required."""


class QMI_NoActiveContextException(QMI_UsageException):
    """Raised when an active default context is required while none exists."""


class QMI_InvalidOperationException(QMI_Exception):
    """Raised when attempting an operation which is not possible in the current state."""


class QMI_ConfigurationException(QMI_Exception):
    """Raised when an error occurs while processing configuration data."""


class QMI_WrongThreadException(QMI_Exception):
    """Raised when a function detects that it is called in a wrong thread."""


#
# Exceptions caused by interaction with instruments/hardware:
#

class QMI_InstrumentException(QMI_Exception):
    """Raised when an interaction with an instrument fails."""


class QMI_TimeoutException(QMI_Exception):
    """Raised when timeout occurs before a requested operation completes."""


class QMI_EndOfInputException(QMI_Exception):
    """Raised when a read operation can not complete because the end of the input is reached."""


class QMI_TransportDescriptorException(QMI_Exception):
    """Raised when an invalid transport descriptor is used."""


#
# Exceptions related to internal QMI functions:
#

class QMI_MessageDeliveryException(QMI_Exception):
    """Raised when a QMI message can not be delivered to its target."""


class QMI_RpcTimeoutException(QMI_Exception):
    """Raised when a remote procedure call fails to complete within the maximum duration."""


class QMI_UnknownRpcException(QMI_Exception):
    """Raised when an undefined RPC request occurs."""


class QMI_SignalSubscriptionException(QMI_Exception):
    """Raised when attempting to subscribe to an unavailable signal."""


#
# Task management related exceptions:
#

class QMI_TaskInitException(QMI_Exception):
    """An error occurred while instantiating the QMI_Task."""


class QMI_TaskRunException(QMI_Exception):
    """An error occurred while running the QMI_Task."""


class QMI_TaskStopException(QMI_Exception):
    """Raised to stop a sleeping task which has received a stop request."""



"""Instrument driver for the stanford research system sim900 module."""

import time

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException


class Sim900(QMI_Instrument):
    """Instrument driver for the stanford research system sim900 module."""

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver.

        Arguments:
            name: Name for this instrument instance.
            transport: Transport descriptor to access the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"port": 5025})
        self._scpi = ScpiProtocol(self._transport)

    @rpc_method
    def open(self) -> None:
        """See base class."""
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        """See base class."""
        super().close()
        self._transport.close()

    @rpc_method
    def input_bytes_waiting(self, port: int) -> int:
        """Input Bytes Waiting

        Query bytes waiting on the input buffer of the specified port.

        Arguments:
            port: port to query number of input bytes waiting.

        Returns:
            The integer number of bytes waiting to be read by the host.

        """
        self._check_is_open()
        self._check_port(port)
        response = self._scpi.ask(f"NINP? {port}", timeout=0.1)
        try:
            return int(response)
        except ValueError as err:
            raise QMI_InstrumentException(
                f"Expected response to be of type int, instead received \"{response}\".") from err

    @rpc_method
    def get_raw_bytes(self, port: int, num_bytes: int) -> bytes:
        """Get Raw Bytes from Port

        The RAWN command retrieves exactly i bytes from the specified port and return them.

        Arguments:
            port: port number to retrieve from.
            num_bytes: amount of bytes to retrieve.

        Returns:
            The queried bytes.
        """
        self._check_is_open()
        self._check_port(port)
        self._scpi.write(f"RAWN? {port},{num_bytes}")
        return self._transport.read(num_bytes, timeout=0.1)

    @rpc_method
    def send_terminated_message(self, port: int, message: str) -> None:
        """Send Terminated Message to Port

        The send terminated message transfers the message followed by the <term>
        sequence to the specified port.

        Arguments:
            port: port number to send terminated message to.
            message: message to send.
        """
        self._check_is_open()
        self._check_port(port)
        self._scpi.write(f"SNDT {port},\"{message}\"")

    @rpc_method
    def ask_module(self, port: int, message: str, delay: float = 0.500) -> str:
        """Ask module for data.

        This is a helper function that simulates the SCPI _ask_ functionality. It is important that this function is
        executed atomically to avoid race conditions.

        Arguments:
            port: port number of module to ask for data
            message: message to send
            delay: how long the function should wait before the full response is available

        Returns:
           The ascii decoded message excluding the terminator.
        """
        self.send_terminated_message(port, message)
        time.sleep(delay)
        num_bytes = self.input_bytes_waiting(port)
        response = self.get_raw_bytes(port, num_bytes)
        if not response.endswith('\n'.encode('ascii')):
            raise QMI_InstrumentException(f"Bad response, expected \"\\n\", received {response[-1]!r}.")
        decoded_response = response[:-1].decode("ascii")
        return decoded_response

    def _check_port(self, port: int):
        """ Validate the port number.

        Checks whether the specified port number is valid for the SIM900.

        Arguments:
            port: port number to validate.

        Raises:
            QMI_UsageException: Invalid port specified
        """
        if not 1 <= port <= 8:
            raise QMI_UsageException(f"Port should be between 1 and 8, but is {port}.")



"""Initialization of the Python logging framework for QMI."""

import sys
import logging
import logging.handlers
import threading
import time
import warnings

from types import TracebackType
from typing import Dict, Callable, Mapping, Optional, Tuple, Type, Union


# Global variable holding the file log handler (if any).
_file_handler: Optional[logging.FileHandler] = None

# Global variable holding the saved exception hook.
_saved_except_hook: Optional[Callable] = None


class _RateLimitFilter(logging.Filter):
    """Rate limiting of log messages.

    Rate limiting can be used to prevent excessive amounts of repeated log messages filling up the disk.

    The rate limits are tracked separately per logger and per log level.
    Conceptually, each combination of logger and log level has an associated `token bucket`.
    When a message is logged, one token is taken out of the bucket.
    The bucket then gets refilled at a steady rate until it is full again.
    When the bucket becomes empty, log messages are dropped until the bucket is sufficiently refilled.
    """

    def __init__(self, rate_limit: float, burst_limit: int = 1) -> None:
        """Initialize a rate limiting log filter.

        Parameters:
            rate_limit: Maximum number of log messages per second.
            burst_limit: Maximum number of messages that can be "saved up" for a short burst of messages.
                         This must be at least 1.
        """
        super().__init__("")
        self._rate_limit = rate_limit
        self._burst_limit = burst_limit
        self._lock = threading.Lock()

        # _counters is used to track the token bucket state for each
        # combination of logger and log level:
        #    self._counters[(logger_name, log_level)] returns a tuple
        #      (last_update_time, bucket_level, nr_discarded_messages)
        self._counters: Dict[Tuple[str, int], Tuple[float, float, int]] = {}

    def filter(self, record: logging.LogRecord) -> bool:

        time_now = time.monotonic()

        with self._lock:
            # Get counter record for this combination of logger and log level.
            key = (record.name, record.levelno)
            counter_data = self._counters.get(key)

            if counter_data is not None:
                # Update the bucket level.
                # Refill the bucket by (rate_limit * elapsed_time) tokens,
                # but never more than the burst limit.
                (last_time, bucket_level, messages_discarded) = counter_data
                bucket_level += self._rate_limit * (time_now - last_time)
                bucket_level = min(bucket_level, self._burst_limit)
            else:
                # First use. Initialize the bucket level to the to burst limit.
                bucket_level = self._burst_limit
                messages_discarded = 0

            if bucket_level < 1:
                # Discard this message.
                messages_discarded += 1
                ret = False
            else:
                # Allow this message.
                bucket_level -= 1.0
                ret = True

            # Store updated counter record.
            self._counters[key] = (time_now, bucket_level, messages_discarded)

        return ret


def _makeLogFormatter(log_process: bool) -> logging.Formatter:
    """Create a log formatter instance."""

    if log_process:
        format_spec = "%(asctime)-23s | %(process)5d | %(levelname)-8s | %(name)-22s | %(message)s"
    else:
        format_spec = "%(asctime)-23s | %(levelname)-8s | %(name)-22s | %(message)s"

    fmt = logging.Formatter(format_spec)

    # Override default timezone and time format.
    fmt.converter = time.gmtime  # type: ignore
    fmt.default_msec_format = "%s.%03d"

    return fmt


def start_logging(loglevel: Union[int, str] = logging.INFO,
                  console_loglevel: Union[int, str] = logging.WARNING,
                  logfile: Optional[str] = None,
                  loglevels: Optional[Mapping[str, Union[int, str]]] = None,
                  rate_limit: Optional[float] = None,
                  burst_limit: int = 1
                  ) -> None:
    """Initialize the Python logging framework for use by QMI.

    This function sets up logging to `stderr` and optional logging to a file.
    Logging of Python *warnings* is enabled.
    Unhandled exceptions are logged to file.

    This function is normally called automatically by ``qmi.start()``.

    Parameters:
        loglevel: Default log level (level of the root logger).
            Only log messages with at least this priority will be processed.
        console_loglevel: Log level for logging to console.
            Only log messages with at least this priority will be logged to screen.
        logfile: Optional file name of the log file.
            When omitted, logging to file will be disabled.
        loglevels: Optional initial log levels for specific loggers.
            When specified, this is a dictionary mapping logger names to
            their initial log levels.
        rate_limit: Maximum number of log messages per second per logger.
        burst_limit: Maximum number of messages that can be "saved up" for a short burst of messages.
    """

    global _file_handler
    global _saved_except_hook

    # If there is still an old file log handler, remove it.
    if _file_handler is not None:
        logging.getLogger().removeHandler(_file_handler)
        _file_handler.close()
        _file_handler = None

    # Set log level of the root logger.
    logging.getLogger().setLevel(loglevel)

    # Set basic configuration: logging to stderr.
    hdlr = logging.StreamHandler()
    fmt = _makeLogFormatter(log_process=False)
    hdlr.setFormatter(fmt)
    hdlr.setLevel(console_loglevel)
    logging.basicConfig(handlers=[hdlr])

    # Create log file handler.
    # This must be done as a separate step (separate from basicConfig())
    # because basicConfig() ignores subsequent calls once logging is configured.
    # However we must support the case where logging is configured during
    # early initialization via QMI_DEBUG, then later re-configured to add
    # a log file after the configuration is processed.
    if logfile:
        # Use the WatchedFileHandler class for logging to file.
        # This handler will automatically re-open the log file if the underlying
        # file is removed or renamed, for example as part of log rotation.
        _file_handler = logging.handlers.WatchedFileHandler(logfile)
        fmt = _makeLogFormatter(log_process=True)
        _file_handler.setFormatter(fmt)
        # Configure rate limiting.
        if rate_limit is not None:
            _file_handler.addFilter(_RateLimitFilter(rate_limit, burst_limit))
        logging.getLogger().addHandler(_file_handler)

    # Set log levels of specific loggers.
    if loglevels:
        for (k, v) in loglevels.items():
            logging.getLogger(k).setLevel(v)

    # Set hook to log uncaught exceptions.
    if _saved_except_hook is None:
        _saved_except_hook = sys.excepthook
        sys.excepthook = _log_excepthook

    # Redirect warnings through the logging system.
    logging.captureWarnings(True)

    # By default the Python interpreter disables many types of warnings.
    # We change that here, unless explicit warning options have been specified via "python -W ...".
    if not sys.warnoptions:
        # Enable all warnings but log only the first occurrence.
        warnings.simplefilter("default")


def _log_excepthook(type_: Type[BaseException],
                    value: BaseException,
                    traceback: Optional[TracebackType] = None
                    ) -> None:
    """Called when an uncaught exception occurs.

    This function logs the exception to the QMI log file (if any),
    then calls the original exception hook to display the exception
    on screen.
    """

    assert traceback is not None
    # Call the original excepthook to display the exception on screen.
    assert _saved_except_hook is not None
    _saved_except_hook(type_, value, traceback)

    # Log the exception to file.
    hdlr = _file_handler
    if hdlr is not None:
        # Create temporary logger which logs only to file (not to screen).
        logger = logging.Logger("exception")
        logger.addHandler(hdlr)
        # Log the exception.
        logger.error("%s: %s", type_.__name__, value, exc_info=(type_, value, traceback))



"""
Thorlabs Inc.; motorized mounts, optical power meter, temperature controller, environmental sensors.

The qmi.instruments.thorlabs package provides support for:
- K10CR1 rotational mount
- MFF10X filter flip mounts
- PM100D power meter
- TSP01, TSP01B environmental sensors.
- MPC320 Polarisation Controller.
"""

from qmi.instruments.thorlabs.pm100d import SensorInfo
from qmi.instruments.thorlabs.tc200 import Tc200Status
from qmi.instruments.thorlabs.apt_protocol import AptChannelState

# Alternative, QMI naming convention approved names
from qmi.instruments.thorlabs.k10cr1 import Thorlabs_K10CR1 as Thorlabs_K10Cr1
from qmi.instruments.thorlabs.mff10x import Thorlabs_MFF10X as Thorlabs_Mff10X
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM16_120 as Thorlabs_pm16120
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM100D as Thorlabs_Pm100d
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM100USB as Thorlabs_Pm100usb
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM101U as Thorlabs_Pm101u
from qmi.instruments.thorlabs.tc200 import Thorlabs_TC200 as Thorlabs_Tc200
from qmi.instruments.thorlabs.tsp01 import Thorlabs_TSP01 as Thorlabs_Tsp01
from qmi.instruments.thorlabs.tsp01b import Thorlabs_TSP01B as Thorlabs_Tsp01b
from qmi.instruments.thorlabs.mpc320 import Thorlabs_Mpc320



"""
JPE, piezo-based cryogenic compatible positioning.

The qmi.instruments.jpe package provides support for:
- Cryo Positioning Systems Controller (CPSC).
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.jpe.cpsc import JPE_CPSC as Jpe_CPSC
from qmi.instruments.jpe.cpsc import StatusPositionControl



#! /usr/bin/env python3

"""Test basic start/stop functionality of QMI framework.
"""
import os
import unittest
from collections import OrderedDict

ORIGINAL_QMI_CONFIG = os.getenv("QMI_CONFIG")
qmi_config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "qmi.conf")
os.environ["QMI_CONFIG"] = qmi_config_path

import qmi
import qmi.core.context_singleton
import qmi.core.exceptions

QMI_ENV_CONFIG_FILE_PRESENT = False
QMI_ENV_CONFIG_FILE = os.path.join(str(os.getenv("QMI_CONFIG")), "qmi.conf")
if os.path.isfile(QMI_ENV_CONFIG_FILE):
    QMI_ENV_CONFIG_FILE_PRESENT = True

CONFIG_FILE_CONTEXT = """{
    # Log level for messages to the console.
    "logging": {
        "console_loglevel": "WARNING"
    },

    "contexts": {
        # Testing remote instrument access.
        "instr_server": {
            "host": "127.0.0.1",
            "tcp_server_port": 40001
        }
    }
}"""


class TestContextConfigFileInputs(unittest.TestCase):

    def tearDown(self):
        # Set these back to None to avoid false leads for later unit-tests
        qmi.core.context_singleton.QMI_CONFIG = None

    def test_01_context_name_in_config_file(self):
        # Check that the context name matches with a config file context name.
        # First create a temporary qmi.conf file.
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")
        with open(qmi_config_file, 'w') as qmi_conf:
            qmi_conf.write(CONFIG_FILE_CONTEXT)

        context_name = "instr_server"

        try:
            qmi.start(context_name, qmi_config_file)
            contexts = qmi.get_configured_contexts()
            self.assertIsInstance(contexts, OrderedDict)
            self.assertTrue(context_name in contexts.keys())

        finally:
            qmi.stop()
            os.remove(qmi_config_file)

    def test_02_context_name_not_in_config_file(self):
        # Check that, when context name is not in the config file, an exception is raised when checking the configured
        # contexts.
        # First create a temporary qmi.conf file.
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")
        with open(qmi_config_file, 'w') as qmi_conf:
            qmi_conf.write(CONFIG_FILE_CONTEXT)

        context_name = "hello_world"

        qmi.start(context_name, qmi_config_file)
        try:
            contexts = qmi.get_configured_contexts()
            self.assertIsInstance(contexts, OrderedDict)
            self.assertTrue(context_name not in contexts.keys())

        finally:
            qmi.stop()
            os.remove(qmi_config_file)

    def test_03_no_config_file(self):
        # Check that, when no config file is defined, an exception is raised when checking the configured contexts.
        context_name = "instr_server"

        try:
            qmi.start(context_name)
            contexts = qmi.get_configured_contexts()
            self.assertIsInstance(contexts, OrderedDict)
            self.assertTrue(context_name not in contexts.keys())

        finally:
            qmi.stop()


if __name__ == "__main__":
    unittest.main()
    os.environ["QMI_CONFIG"] = ORIGINAL_QMI_CONFIG



"""Unit test for Newport/NewFocus TLB670-X."""
import logging
import unittest
from unittest.mock import patch, Mock

from qmi.core.exceptions import QMI_InstrumentException
from qmi.instruments.newport.tlb670x import NewFocus_TLB670X


class CtypesMock:
    """Mock ctypes library.

    User create_string_buffer() and byref() as follows:
    1. push a StringBuffer object to the string_buffer attribute
    2. create_string_buffer() pops from the same attribute
    3. byref() adds it argument on the ref_objs attribute: this should be the same object that you added in step 1; the
       id() of that object is returned by byref() so you can check for that in the call arguments of the dllmock calls.
    """
    class StringBuffer:
        def __init__(self, value: str):
            self.value = value.encode("ascii")

        def __repr__(self):
            return "StringBuffer(id={}, value={})".format(hex(id(self)), self.value.decode())

    class IntValue:
        def __init__(self, value):
            self.value = value

        def __repr__(self):
            return "IntValue(id={}, value={})".format(hex(id(self)), self.value)

    def __init__(self):
        self.dllmock = Mock()
        self.ref_objs = []
        self.string_buffer = []

    def WinDLL(self, _):
        return self.dllmock

    def byref(self, obj):
        self.ref_objs.append(obj)
        return id(obj)

    def create_string_buffer(self, _):
        return self.string_buffer.pop(0)

    def c_ulong(self, value):
        return self.IntValue(value)


class TestTLB670XPlatform(unittest.TestCase):
    def setUp(self):
        patcher = patch("qmi.instruments.newport.tlb670x.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.configure_mock(platform="linux")
        self.addCleanup(patcher.stop)

        self.ctypes_mock = CtypesMock()
        patcher = patch("qmi.instruments.newport.tlb670x.ctypes", self.ctypes_mock)
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_platform_support(self):
        """Test platform support (Windows only)."""
        with self.assertRaises(RuntimeError):
            NewFocus_TLB670X(Mock(), "tlb670x", "anything")


class TestTLB670XInit(unittest.TestCase):
    def setUp(self):
        patcher = patch("qmi.instruments.newport.tlb670x.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.platform = "win"
        self.addCleanup(patcher.stop)

        self.ctypes_mock = CtypesMock()
        patcher = patch("qmi.instruments.newport.tlb670x.ctypes", self.ctypes_mock)
        patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.NOTSET)

    def test_init_nominal(self):
        """Test nominal initialization."""
        serial_number = "SN1234"
        device_id = 123
        device_info_response = CtypesMock.StringBuffer(f"1,SNxxxx;2,SNyyyy;{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        instr.open()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)
        self.assertEqual(instr.get_device_id(), device_id)
        self.assertEqual(instr.get_ident().serial, serial_number)

    def test_init_with_numeric_serialno(self):
        """Test with numeric serial number."""
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.CRITICAL)

        serial_number = "SN1234"
        device_id = 123
        device_info_response = CtypesMock.StringBuffer(f"1,SNxxxx;2,SNyyyy;{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", 1234)
        instr.open()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)
        self.assertEqual(instr.get_device_id(), device_id)
        self.assertEqual(instr.get_ident().serial, serial_number)

    def test_init_with_serialno_no_prefix(self):
        """Test with serial number without SN prefix."""
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.CRITICAL)

        serial_number = "SN1234"
        device_id = 123
        device_info_response = CtypesMock.StringBuffer(f"1,SNxxxx;2,SNyyyy;{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", "1234")
        instr.open()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)
        self.assertEqual(instr.get_device_id(), device_id)
        self.assertEqual(instr.get_ident().serial, serial_number)

    def test_init_device_not_found(self):
        """Test no device present."""
        serial_number = "SN1234"
        device_info_response = CtypesMock.StringBuffer("1,SNxxxx;2,SNyyyy;")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        with self.assertRaises(QMI_InstrumentException):
            instr.open()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)

    def test_init_no_devices(self):
        """Test no device present."""
        serial_number = "SN1234"
        device_info_response = CtypesMock.StringBuffer("")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        with self.assertRaises(QMI_InstrumentException):
            instr.open()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)

    def test_init_deinit(self):
        """Test init/deinit."""
        serial_number = "SN1234"
        device_id = 123
        device_info_response = CtypesMock.StringBuffer(f"1,SNxxxx;2,SNyyyy;{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        instr.open()
        instr.close()

        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once_with(id(device_info_response))
        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs.pop(), device_info_response)


class TestTLB670XTxRx(unittest.TestCase):
    def setUp(self):
        patcher = patch("qmi.instruments.newport.tlb670x.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.platform = "win"
        self.addCleanup(patcher.stop)

        self.ctypes_mock = CtypesMock()
        patcher = patch("qmi.instruments.newport.tlb670x.ctypes", self.ctypes_mock)
        patcher.start()
        self.addCleanup(patcher.stop)

        serial_number = "SN1234"
        device_id = 123
        device_info_response = CtypesMock.StringBuffer(f"{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        self.instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        self.instr.open()

        self.ctypes_mock.ref_objs.pop(0)  # call to get_device_info
        self.ctypes_mock.dllmock.reset_mock()

    def tearDown(self):
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.NOTSET)

    def test_txrx_nominal(self):
        """Test nominal send/receive sequence."""
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        error_str = "my_error_string"
        error_str_response = CtypesMock.StringBuffer(error_str)
        self.ctypes_mock.string_buffer.append(error_str_response)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        self.instr.check_error_status()

        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, "ERRSTR?".encode("ascii"))  # query
        self.assertEqual(self.ctypes_mock.ref_objs[1].value, error_str.encode("ascii"))  # response

    def test_tx_fail(self):
        """Test failure in send sequence."""
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.CRITICAL)

        expected_error = "Command ERRSTR? returned an error -2: COMMAND NOT VALID"
        # Setup for _send
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        error_code = -2  # COMMAND NOT VALID
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = error_code

        # Setup for reinit (automatic after fail in send)
        device_info_response = CtypesMock.StringBuffer("123,SN1234")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.check_error_status()
            self.assertEqual(expected_error, str(exc.exception))

        # Check send/receive sequence
        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_not_called()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, "ERRSTR?".encode("ascii"))  # query

        # Check for reinit
        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()

    def test_rx_fail(self):
        """Test failure in receive sequence."""
        logging.getLogger("qmi.instruments.newport.tlb670x").setLevel(logging.CRITICAL)

        # Setup for _send
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0

        # Setup for _receive
        error_code = -2  # COMMAND NOT VALID
        error_str = "my_error_string"
        error_str_response = CtypesMock.StringBuffer(error_str)
        self.ctypes_mock.string_buffer.append(error_str_response)  # need to add for call to get_ascii
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = error_code

        # Setup for reinit (automatic after fail in receive)
        device_info_response = CtypesMock.StringBuffer("123,SN1234")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        with self.assertRaises(QMI_InstrumentException):
            self.instr.check_error_status()

        # Check send/receive sequence
        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, "ERRSTR?".encode("ascii"))  # query

        # Check for reinit
        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()

    def test_reset(self):
        """Test 'reset' RPC call."""
        self.instr.RESET_SLEEP_TIME = 0.01
        error_str = "NO ERROR"
        command = CtypesMock.StringBuffer("ERRSTR?")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        error_str_response = CtypesMock.StringBuffer(error_str)
        self.ctypes_mock.string_buffer.append(error_str_response)
        command = CtypesMock.StringBuffer("*RST")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        ok_response = CtypesMock.StringBuffer("OK\r\n")
        self.ctypes_mock.string_buffer.append(ok_response)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        self.instr.reset()

        self.assertEqual(2, self.ctypes_mock.dllmock.newp_usb_send_ascii.call_count)
        self.assertEqual(2, self.ctypes_mock.dllmock.newp_usb_get_ascii.call_count)

    def test_reset_excepts(self):
        """Test 'reset' RPC call with raising QMI_InstrumentException."""
        self.instr.RESET_SLEEP_TIME = 0.01
        error_str = "NO ERROR"
        command = CtypesMock.StringBuffer("ERRSTR?")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        error_str_response = CtypesMock.StringBuffer(error_str)
        self.ctypes_mock.string_buffer.append(error_str_response)
        command = CtypesMock.StringBuffer("*RST")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        serial_number = "SN1234"
        device_id = 12
        device_info_response = CtypesMock.StringBuffer(f"{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.side_effect = [0, 116]
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        with self.assertRaises(QMI_InstrumentException):
            self.instr.reset()

        self.ctypes_mock.dllmock.newp_usb_uninit_system.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_init_product.assert_called_once_with(NewFocus_TLB670X.PRODUCT_ID)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()
        self.assertEqual(2, self.ctypes_mock.dllmock.newp_usb_send_ascii.call_count)


class TestTLB670XGetSet(unittest.TestCase):
    def setUp(self):
        patcher = patch("qmi.instruments.newport.tlb670x.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.platform = "win"
        self.addCleanup(patcher.stop)

        self.ctypes_mock = CtypesMock()
        patcher = patch("qmi.instruments.newport.tlb670x.ctypes", self.ctypes_mock)
        patcher.start()
        self.addCleanup(patcher.stop)

        serial_number = "SN1234"
        self.device_id = 12
        device_info_response = CtypesMock.StringBuffer(f"{self.device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        self.instr = NewFocus_TLB670X(Mock(), "tlb670x", serial_number)
        self.instr.open()

        self.ctypes_mock.ref_objs.pop(0)  # call to get_device_info
        self.ctypes_mock.dllmock.reset_mock()

    def _generic_get_test(self, query, expected_value, getter):
        """Helper function for testing getters."""
        # Setup send/receive
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        value_str = str(expected_value) + "\r\n"
        response = CtypesMock.StringBuffer(value_str)
        self.ctypes_mock.string_buffer.append(response)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        return_value = getter()

        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, query.encode("ascii"))  # query
        self.assertEqual(self.ctypes_mock.ref_objs[1].value, value_str.encode("ascii"))  # response
        self.assertEqual(return_value, expected_value)

    def _generic_get_test_with_extra_OKs(self, query, expected_value, getter, extra_oks=1, after_oks=0, second_read=""):
        """Helper function for testing getters, return extra "OK" before correct answer."""
        # Setup send/receive
        command = CtypesMock.StringBuffer("4\r\n;4\r\n")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        extra_oks_string = ""
        for _ in range(extra_oks):
            extra_oks_string += "OK\r\n"

        after_oks_string = ""
        for _ in range(after_oks):
            after_oks_string += "OK\r\n"

        value_str = extra_oks_string + str(expected_value) + "\r\n" + after_oks_string
        response = CtypesMock.StringBuffer(value_str)
        self.ctypes_mock.string_buffer.append(response)
        if second_read:
            self.ctypes_mock.string_buffer.append(CtypesMock.StringBuffer(second_read))

        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        return_value = getter()

        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        if second_read:
            self.assertEqual(2, self.ctypes_mock.dllmock.newp_usb_get_ascii.call_count)
        else:
            self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()

        self.assertEqual(self.ctypes_mock.ref_objs[0].value, query.encode("ascii"))  # query
        self.assertEqual(self.ctypes_mock.ref_objs[1].value, value_str.encode("ascii"))  # response
        self.assertEqual(return_value, expected_value)

    def _generic_get_test_with_extra_IDN(self, query, expected_value, getter):
        """Helper function for testing getters, return extra "IDN" after correct answer."""
        # Setup send/receive
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        extra_idn_string = "cus TLB-6700 v2.4 31/09/23 SN12345\r\n"

        value_str = str(expected_value) + "\r\n" + extra_idn_string
        response = CtypesMock.StringBuffer(value_str)
        self.ctypes_mock.string_buffer.append(response)

        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        return_value = getter()

        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, query.encode("ascii"))  # query
        self.assertEqual(self.ctypes_mock.ref_objs[1].value, value_str.encode("ascii"))  # response
        self.assertEqual(return_value, expected_value)

    def _generic_set_test(self, cmd_str, setpoint_value, setter):
        """Helper function for testing getters."""
        # Setup send/receive
        command = CtypesMock.StringBuffer("")  # to be set by _send
        self.ctypes_mock.string_buffer.append(command)
        response = CtypesMock.StringBuffer("OK\r\n")
        self.ctypes_mock.string_buffer.append(response)
        self.ctypes_mock.dllmock.newp_usb_send_ascii.return_value = 0  # success
        self.ctypes_mock.dllmock.newp_usb_get_ascii.return_value = 0  # success

        setter(setpoint_value)

        self.ctypes_mock.dllmock.newp_usb_send_ascii.assert_called_once()
        self.ctypes_mock.dllmock.newp_usb_get_ascii.assert_called_once()
        self.assertEqual(self.ctypes_mock.ref_objs[0].value, cmd_str.format(setpoint_value).encode("ascii"))  # command

    def test_get_available_devices_info(self):
        """Test getting available device info."""
        serial_number = "SN1234"
        device_id = 12
        expected_info = [(device_id, serial_number)]
        device_info_response = CtypesMock.StringBuffer(f"{device_id},{serial_number};")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = 0  # success

        info = self.instr.get_available_devices_info()

        self.assertListEqual(expected_info, info)

    def test_get_available_devices_info_excepts(self):
        """Test the exception case for getting devices info excepts at re-init."""
        expected_exception = "Unable to load device info: Unknown error"
        device_info_response = CtypesMock.StringBuffer("None,None;")
        self.ctypes_mock.string_buffer.append(device_info_response)
        device_info_response_reinit = CtypesMock.StringBuffer(f"{self.device_id},SOMESN;")
        self.ctypes_mock.string_buffer.append(device_info_response_reinit)
        self.ctypes_mock.dllmock.newp_usb_get_device_info.return_value = -1  # Some failure

        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.get_available_devices_info()

        self.assertEqual(expected_exception, str(exc.exception))

    def test_get_available_devices_info_excepts_reinit(self):
        """Test the exception case for getting devices info excepts at re-init."""
        expected_exception = "No TLB-670X instrument present"
        device_info_response = CtypesMock.StringBuffer("None,None;")
        self.ctypes_mock.string_buffer.append(device_info_response)
        self.ctypes_mock.string_buffer.append(CtypesMock.StringBuffer(""))
        self.ctypes_mock.dllmock.newp_usb_get_device_info.side_effect = [-1, 0]  # Some failure, then ok

        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.get_available_devices_info()

        self.assertEqual(expected_exception, str(exc.exception))

    def test_get_wavelength(self):
        """Test get wavelength."""
        self._generic_get_test("SOURce:WAVElength?", 123.456, self.instr.get_wavelength)

    def test_get_powermode(self):
        """Test get powermode."""
        self._generic_get_test("OUTPut:STATe?", 1, self.instr.get_powermode)

    def test_get_trackingmode(self):
        """Test get trackingmode."""
        self._generic_get_test("OUTPut:TRACk?", 0, self.instr.get_trackingmode)

    def test_get_diode_current(self):
        """Test get diode current."""
        self._generic_get_test("SENSe:CURRent:DIODe", 456.789, self.instr.get_diode_current)

    def test_get_piezo_voltage(self):
        """Test get piezo voltage."""
        self._generic_get_test("SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage)

    def test_get_piezo_voltage_ok(self):
        """Test get piezo voltage with one extra "OK" returned first."""
        self._generic_get_test_with_extra_OKs("SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage)

    def test_get_piezo_voltage_idn(self):
        """Test get piezo voltage with one extra *IDN response returned after query."""
        self._generic_get_test_with_extra_IDN("SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage)

    def test_get_piezo_voltage_ok_nok(self):
        """Test get piezo voltage with two extra "OK"s returned first. This result in """
        self._generic_get_test_with_extra_OKs("SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage, 2)

    def test_get_piezo_voltage_nok(self):
        """Test get piezo voltage fails if after the expected value there is one "OK" more AFTER the response.
        In this case, as there is one OK
        """
        self._generic_get_test_with_extra_OKs(
                "SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage, 0, 1
            )

    def test_get_piezo_voltage_2nd_read_ok(self):
        """Test get piezo voltage fails if after the expected value there is one "OK" more AFTER the response.
        This case fails because response[1] is also "OK", prompting a re-read which then succeeds."""
        self._generic_get_test_with_extra_OKs(
                "SOURce:VOLTage:PIEZo?", 99.99, self.instr.get_piezo_voltage, 1, 1, "99.99\r\n"
            )

    def test_set_wavelength(self):
        """Test set wavelength."""
        self._generic_set_test("SOURce:WAVElength {}", 1.234, self.instr.set_wavelength)

        with self.assertRaises(ValueError):
            self.instr.set_wavelength(-1.0)

    def test_set_powermode(self):
        """Test set powermode."""
        self._generic_set_test("OUTPut:STATe {}", 0, self.instr.set_powermode)

        with self.assertRaises(ValueError):
            self.instr.set_powermode(-1)

        with self.assertRaises(ValueError):
            self.instr.set_powermode(2)

    def test_set_trackingmode(self):
        """Test set trackingmode."""
        self._generic_set_test("OUTPut:TRACk {}", 0, self.instr.set_trackingmode)

        with self.assertRaises(ValueError):
            self.instr.set_trackingmode(-1)

        with self.assertRaises(ValueError):
            self.instr.set_trackingmode(2)

    def test_set_diode_current_float(self):
        """Test set diode current."""
        self._generic_set_test("SOURce:CURRent:DIODe {}", 123.45, self.instr.set_diode_current)

        with self.assertRaises(ValueError):
            self.instr.set_diode_current(-1.234)

    def test_set_diode_current_int(self):
        """Test set diode current."""
        self._generic_set_test("SOURce:CURRent:DIODe {}", 123, self.instr.set_diode_current)

        with self.assertRaises(ValueError):
            self.instr.set_diode_current(-123)

    def test_set_diode_current_str(self):
        """Test set diode current."""
        self._generic_set_test("SOURce:CURRent:DIODe {}", "MAX", self.instr.set_diode_current)

        with self.assertRaises(ValueError):
            self.instr.set_diode_current("MIN")

    def test_set_piezo_voltage(self):
        """Test set diode current."""
        self._generic_set_test("SOURce:VOLTage:PIEZo {}", 99.99, self.instr.set_piezo_voltage)

        with self.assertRaises(ValueError):
            self.instr.set_piezo_voltage(-1.00)

        with self.assertRaises(ValueError):
            self.instr.set_piezo_voltage(101.00)


if __name__ == '__main__':
    unittest.main()
